use serde::{Deserialize, Serialize};

use crate::{
    Either,
    block_state::BlockState,
    color::{ARGBu8, RGBAf32, RGBf32, RGBu8},
    identifier::Identifier,
    item_stack::ItemStack,
};

#[derive(Debug, Serialize, Deserialize, Clone)]
#[serde(tag = "type")]
pub enum Particle {
    #[serde(rename = "minecraft:angry_villager")]
    AngryVillager,
    #[serde(rename = "minecraft:ash")]
    Ash,
    Block {
        block_state: Either<Identifier, BlockState>,
    },
    #[serde(rename = "minecraft:block_crumble")]
    BlockCrumble {
        block_state: Either<Identifier, BlockState>,
    },
    BlockMarker {
        block_state: Either<Identifier, BlockState>,
    },
    #[serde(rename = "minecraft:bubble")]
    Bubble,
    #[serde(rename = "minecraft:bubble_column_up")]
    BubbleColumnUp,
    #[serde(rename = "minecraft:bubble_pop")]
    BuublePop,
    #[serde(rename = "minecraft:campfire_cosy_smoke")]
    CampfireCosySmoke,
    #[serde(rename = "minecraft:campfire_signal_smoke")]
    CampfireSignalSmoke,
    #[serde(rename = "minecraft:cherry_leaves")]
    CherryLeaves,
    #[serde(rename = "minecraft:cloud")]
    Cloud,
    #[serde(rename = "minecraft:composter")]
    Composter,
    #[serde(rename = "minecraft:crimson_spore")]
    CrimsonSpore,
    #[serde(rename = "minecraft:crit")]
    Crit,
    #[serde(rename = "minecraft:current_down")]
    CurrentDown,
    #[serde(rename = "minecraft:damage_indicator")]
    DamageIndicator,
    #[serde(rename = "minecraft:dolphin")]
    Dolphin,
    #[serde(rename = "minecraft:dragon_breath")]
    DragonBreath,
    #[serde(rename = "minecraft:dripping_dripstone_lava")]
    DrippingDripstoneLava,
    #[serde(rename = "minecraft:dripping_dripstone_water")]
    DrippingDripstoneWater,
    #[serde(rename = "minecraft:dripping_honey")]
    DrippingHoney,
    #[serde(rename = "minecraft:dripping_lava")]
    DrippingLava,
    #[serde(rename = "minecraft:dripping_obsidian_tear")]
    DrippingObsidianTear,
    #[serde(rename = "minecraft:dripping_water")]
    DrippingWater,
    Dust {
        color: RGBf32,
        scale: f32,
    },
    DustColorTransition {
        from_color: RGBf32,
        to_color: RGBf32,
        scale: f32,
    },
    DustPillar {
        block_state: Either<Identifier, BlockState>,
    },
    #[serde(rename = "minecraft:dust_plume")]
    DustPlume,
    #[serde(rename = "minecraft:effect")]
    Effect,
    #[serde(rename = "minecraft:egg_crack")]
    EggCrack,
    #[serde(rename = "minecraft:elder_guardian")]
    ElderGuardian,
    #[serde(rename = "minecraft:electric_spark")]
    ElectricSpark,
    #[serde(rename = "minecraft:enchant")]
    Enchant,
    #[serde(rename = "minecraft:enchanted_hit")]
    EnchantedHit,
    #[serde(rename = "minecraft:end_rod")]
    EndRod,
    EntityEffect {
        color: Either<RGBAf32, ARGBu8>,
    },
    #[serde(rename = "minecraft:explosion")]
    Explosion,
    #[serde(rename = "minecraft:explosion_emitter")]
    ExplosionEmitter,
    #[serde(rename = "minecraft:falling_dripstone_lava")]
    FallingDripstoneLava,
    #[serde(rename = "minecraft:falling_dripstone_water")]
    FallingDrpistoneWater,
    FallingDust {
        block_state: Either<Identifier, BlockState>,
    },
    #[serde(rename = "minecraft:falling_honey")]
    FallingHoney,
    #[serde(rename = "minecraft:falling_lava")]
    FallingLava,
    #[serde(rename = "minecraft:falling_nectar")]
    FallingNectar,
    #[serde(rename = "minecraft:falling_obsidian_tear")]
    FallingObsidianTear,
    #[serde(rename = "minecraft:falling_spore_blossom")]
    FallingSporeBlossom,
    #[serde(rename = "minecraft:falling_water")]
    FallingWater,
    // Firefly, (Upcoming: 1.21.5)
    #[serde(rename = "minecraft:firework")]
    Firework,
    #[serde(rename = "minecraft:fishing")]
    Fishing,
    #[serde(rename = "minecraft:flame")]
    Flame,
    #[serde(rename = "minecraft:flash")]
    Flash,
    #[serde(rename = "minecraft:glow")]
    Glow,
    #[serde(rename = "minecraft:glow_squid_ink")]
    GlowSquidInk,
    #[serde(rename = "minecraft:gust")]
    Gust,
    #[serde(rename = "minecraft:gust_emitter")]
    GustEmitter,
    #[serde(rename = "minecraft:happy_villager")]
    HappyVillager,
    #[serde(rename = "minecraft:heart")]
    Heart,
    #[serde(rename = "minecraft:infested")]
    Infested,
    #[serde(rename = "minecraft:instant_effect")]
    InstantEffect,
    Item {
        item: Either<Identifier, ItemStack>,
    },
    #[serde(rename = "minecraft:item_cobweb")]
    ItemCobweb,
    #[serde(rename = "minecraft:item_slime")]
    ItemSlime,
    #[serde(rename = "minecraft:item_snowball")]
    ItemSnowball,
    #[serde(rename = "minecraft:landing_honey")]
    LandingHoney,
    #[serde(rename = "minecraft:landing_lava")]
    LandingLava,
    #[serde(rename = "minecraft:landing_obsidian_tear")]
    LandingObsidianTear,
    #[serde(rename = "minecraft:large_smoke")]
    LargeSmoke,
    #[serde(rename = "minecraft:lava")]
    Lava,
    #[serde(rename = "minecraft:mycelium")]
    Mycelium,
    #[serde(rename = "minecraft:nautilus")]
    Nautlius,
    #[serde(rename = "minecraft:note")]
    Note,
    #[serde(rename = "minecraft:ominous_spawning")]
    OminousSpawning,
    #[serde(rename = "minecraft:pale_oak_leaves")]
    PaleOakLeaves,
    #[serde(rename = "minecraft:poof")]
    Poof,
    #[serde(rename = "minecraft:portal")]
    Portal,
    #[serde(rename = "minecraft:raid_omen")]
    RaidOmen,
    #[serde(rename = "minecraft:rain")]
    Rain,
    #[serde(rename = "minecraft:reverse_portal")]
    ReversePortal,
    #[serde(rename = "minecraft:scrape")]
    Scrape,
    SculkCharge {
        roll: f32,
    },
    #[serde(rename = "minecraft:sculk_charge_pop")]
    SculkChargePop,
    #[serde(rename = "minecraft:sculk_soul")]
    SculkSoul,
    Shriek {
        delay: i32,
    },
    #[serde(rename = "minecraft:small_flame")]
    SmallFlame,
    #[serde(rename = "minecraft:small_gust")]
    SmallGust,
    #[serde(rename = "minecraft:smoke")]
    Smoke,
    #[serde(rename = "minecraft:sneeze")]
    Sneeze,
    #[serde(rename = "minecraft:snowflake")]
    Snowflake,
    #[serde(rename = "minecraft:sonic_boom")]
    SonicBoom,
    #[serde(rename = "minecraft:soul")]
    Soul,
    #[serde(rename = "minecraft:soul_fire_flame")]
    SoulFireFlame,
    #[serde(rename = "minecraft:spit")]
    Spit,
    #[serde(rename = "minecraft:splash")]
    Splash,
    #[serde(rename = "minecraft:spore_blossom_air")]
    SporeBlossomAir,
    #[serde(rename = "minecraft:squid_ink")]
    SquidInk,
    #[serde(rename = "minecraft:sweep_attack")]
    SweepAttack,
    #[serde(rename = "minecraft:totem_of_undying")]
    TotemOfUndying,
    Trail {
        target: (f32, f32, f32),
        colr: RGBu8,
    },
    #[serde(rename = "minecraft:trail_omen")]
    TrailOmen,
    #[serde(rename = "minecraft:trail_spawner_detection")]
    TrailSpawnerDetection,
    #[serde(rename = "minecraft:underwater")]
    Underwater,
    #[serde(rename = "minecraft:vault_connection")]
    VaultConnection,
    Vibration {
        destination: PositionSource,
        arrival_in_ticks: i32,
    },
    #[serde(rename = "minecraft:warped_spore")]
    WarpedSpore,
    #[serde(rename = "minecraft:wax_off")]
    WaxOff,
    #[serde(rename = "minecraft:wax_on")]
    WaxOn,
    #[serde(rename = "minecraft:white_ash")]
    WhiteAsh,
    #[serde(rename = "minecraft:white_smoke")]
    WhiteSmoke,
    #[serde(rename = "minecraft:witch")]
    Witch,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
#[serde(tag = "type", rename_all = "snake_case")]
pub enum PositionSource {
    Block {
        pos: (i32, i32, i32),
    },
    Entity {
        // source_entity stores a UUID as 4xi32
        source_entity: (i32, i32, i32, i32),
        // y_offset is optional and defaults to 0.0
        y_offset: f32,
    },
}
